# Filename: root.py
# pylint: disable=C0103,R0903
# vim:set ts=4 sts=4 sw=4 et:
"""
Read and write Vanilla ROOT files.
"""

import numpy as np

from km3pipe.logger import get_logger

log = get_logger(__name__)    # pylint: disable=C0103

__author__ = "Moritz Lotze"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal and Moritz Lotze"
__email__ = "tgal@km3net.de"
__status__ = "Development"


def open_rfile(rfile, fmode='r'):
    from rootpy.io import root_open
    if isinstance(rfile, str):
        return root_open(rfile, mode=fmode)
    return rfile


def get_hist(rfile, histname, get_overflow=False):
    """Read a 1D Histogram."""
    import root_numpy as rnp

    rfile = open_rfile(rfile)
    hist = rfile[histname]
    xlims = np.array(list(hist.xedges()))
    bin_values = rnp.hist2array(hist, include_overflow=get_overflow)
    rfile.close()
    return bin_values, xlims


def get_hist2d(rfile, histname, get_overflow=False):
    """Read a 2D Histogram."""
    import root_numpy as rnp

    rfile = open_rfile(rfile)
    hist = rfile[histname]
    xlims = np.array(list(hist.xedges()))
    ylims = np.array(list(hist.yedges()))
    bin_values = rnp.hist2array(hist, include_overflow=get_overflow)
    rfile.close()
    return bin_values, xlims, ylims


def get_hist3d(rfile, histname, get_overflow=False):
    """Read a 3D Histogram."""
    import root_numpy as rnp

    rfile = open_rfile(rfile)
    hist = rfile[histname]
    xlims = np.array(list(hist.xedges()))
    ylims = np.array(list(hist.yedges()))
    zlims = np.array(list(hist.zedges()))
    bin_values = rnp.hist2array(hist, include_overflow=get_overflow)
    rfile.close()
    return bin_values, xlims, ylims, zlims


def interpol_hist2d(h2d, oversamp_factor=10):
    """Sample the interpolator of a root 2d hist.

    Root's hist2d has a weird internal interpolation routine,
    also using neighbouring bins.
    """
    from rootpy import ROOTError

    xlim = h2d.bins(axis=0)
    ylim = h2d.bins(axis=1)
    xn = h2d.nbins(0)
    yn = h2d.nbins(1)
    x = np.linspace(xlim[0], xlim[1], xn * oversamp_factor)
    y = np.linspace(ylim[0], ylim[1], yn * oversamp_factor)
    mat = np.zeros((xn, yn))
    for xi in range(xn):
        for yi in range(yn):
            try:
                mat[xi, yi] = h2d.interpolate(x[xi], y[yi])
            except ROOTError:
                continue
    return mat, x, y



# Filename: cmd.py
"""
Convert HDF5 to vanilla ROOT.

Can convert multiple files at once, "foo.h5" -> "foo.h5.root".

Usage:
     hdf2root [--verbose] FILES...
     hdf2root (-h | --help)
     hdf2root --version

Options:
    -h --help           Show this screen.
    --verbose           Print more info [default: False]
"""

import numpy as np

from km3pipe import version

__author__ = "Moritz Lotze"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal and Moritz Lotze"
__email__ = "tgal@km3net.de"
__status__ = "Development"


def hdf2root(infile, outfile, verbose=False):
    try:
        from rootpy.io import root_open
        from rootpy import asrootpy
        from root_numpy import array2tree
    except ImportError:
        raise ImportError(
            "Please load ROOT into PYTHONPATH and install rootpy+root_numpy:\n"
            "   `pip install rootpy root_numpy`"
        )

    from tables import open_file

    h5 = open_file(infile, 'r')
    rf = root_open(outfile, 'recreate')

    # 'walk_nodes' does not allow to check if is a group or leaf
    #   exception handling is bugged
    #   introspection/typecheck is buged
    # => this moronic nested loop instead of simple `walk`
    for group in h5.walk_groups():
        for leafname, leaf in group._v_leaves.items():
            arr = leaf[:]
            if arr.dtype.names is None:
                dt = np.dtype((arr.dtype, [(leafname, arr.dtype)]))
                arr = arr.view(dt)
            treename = leaf._v_pathname.replace('/', '_')
            tree = asrootpy(array2tree(arr, name=treename))
            tree.write()
    rf.close()
    h5.close()


def main():
    from docopt import docopt
    args = docopt(__doc__, version=version)

    files = args['FILES']
    verbose = bool(args['--verbose'])
    for fn in files:
        print('Converting {}...'.format(fn))
        hdf2root(fn, fn + '.root', verbose=verbose)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Guess the distribution!
=======================

Fit several distributions to angular residuals.

Note: to fit the landau distribution, you need to have ROOT and the
``rootpy`` package installed.

"""

import h5py
import matplotlib.pyplot as plt
import numpy as np
from scipy import stats
from statsmodels.distributions.empirical_distribution import ECDF
from statsmodels.nonparametric.kde import KDEUnivariate

try:
    import ROOT
    import rootpy.plotting
    HAS_ROOT = True
except ImportError:
    HAS_ROOT = False

from km3pipe.stats import bootstrap_fit
import km3pipe.style.moritz    # noqa

##################################################################

DATA_FILE = '../data/residuals.h5'

with h5py.File(DATA_FILE) as h5:
    resid = h5['/residuals'][:]

##################################################################
# Fit somedistributions, and obtain the confidence intervals on the
# distribution parameters through bootstrapping.

n_bs = 5
q = 95

ln_par, ln_lo, ln_up = bootstrap_fit(
    stats.lognorm, resid, n_iter=n_bs, quant=q
)
hc_par, hc_lo, hc_up = bootstrap_fit(
    stats.halfcauchy, resid, n_iter=n_bs, quant=q
)
gam_par, gam_lo, gam_up = bootstrap_fit(
    stats.gamma, resid, n_iter=n_bs, quant=q
)

##################################################################

hc = stats.halfcauchy(*stats.halfcauchy.fit(resid))
lg = stats.lognorm(*stats.lognorm.fit(resid))
dens = KDEUnivariate(resid)
dens.fit()
ecdf = ECDF(resid)

##################################################################
# prepare X axes for plotting

ex = ecdf.x
x = np.linspace(min(resid), max(resid), 2000)

##################################################################
# Fit a Landau distribution with ROOT

if HAS_ROOT:
    root_hist = rootpy.plotting.Hist(100, 0, np.pi)
    root_hist.fill_array(resid)
    root_hist /= root_hist.Integral()

    land_f = ROOT.TF1('land_f', "TMath::Landau(x, [0], [1], 0)")
    fr = root_hist.fit('land_f', "S").Get()
    try:
        p = fr.GetParams()
        land = np.array([ROOT.TMath.Landau(xi, p[0], p[1], True) for xi in x])
        land_cdf = np.array([
            ROOT.ROOT.Math.landau_cdf(k, p[0], p[1]) for k in ex
        ])
    except AttributeError:
        # wtf this fails sometimes, idk, works on root6
        HAS_ROOT = False

##################################################################
# ... and plot everything.

fig, axes = plt.subplots(ncols=2, nrows=2, figsize=(6 * 2, 4 * 2))

axes[0, 0].hist(resid, bins='auto', density=True)
axes[0, 0].plot(x, lg.pdf(x), label='Log Norm')
axes[0, 0].plot(x, hc.pdf(x), label='Half Cauchy')
if HAS_ROOT:
    axes[0, 0].plot(x, land, label='Landau', color='blue')
axes[0, 0].plot(x, dens.evaluate(x), label='KDE')
axes[0, 0].set_xlabel('x')
axes[0, 0].set_xlim(0, 0.3)
axes[0, 0].set_ylabel('PDF(x)')
axes[0, 0].legend()

axes[0, 1].hist(resid, bins='auto', density=True)
axes[0, 1].plot(x, lg.pdf(x), label='Log Norm')
axes[0, 1].plot(x, hc.pdf(x), label='Half Cauchy')
if HAS_ROOT:
    axes[0, 1].plot(x, land, label='Landau', color='blue')
axes[0, 1].plot(x, dens.evaluate(x), label='KDE')
axes[0, 1].set_xlabel('x')
axes[0, 1].set_ylabel('PDF(x)')
axes[0, 1].set_yscale('log')
axes[0, 1].legend()

axes[1, 0].plot(ex, 1 - lg.cdf(ex), label='Log Norm')
if HAS_ROOT:
    axes[1, 0].plot(ex, 1 - land_cdf, label='Landau', color='blue')
axes[1, 0].plot(ex, 1 - hc.cdf(ex), label='Half Cauchy')
axes[1, 0].plot(
    ex, 1 - ecdf.y, label='Empirical CDF', linewidth=3, linestyle='--'
)
axes[1, 0].set_xscale('log')
axes[1, 0].set_xlabel('x')
axes[1, 0].set_ylabel('1 - CDF(x)')
axes[1, 0].legend()

axes[1, 1].loglog(ex, 1 - lg.cdf(ex), label='Log Norm')
if HAS_ROOT:
    axes[1, 1].loglog(ex, 1 - land_cdf, label='Landau', color='blue')
axes[1, 1].loglog(ex, 1 - hc.cdf(ex), label='Half Cauchy')
axes[1, 1].loglog(
    ex, 1 - ecdf.y, label='Empirical CDF', linewidth=3, linestyle='--'
)
axes[1, 1].set_xlabel('x')
axes[1, 1].set_ylabel('1 - CDF(x)')
axes[1, 1].legend()



#!/usr/bin/env python
import time

from km3pipe import Pipeline, Module
from km3pipe.io import CHPump


class CHPrinter(Module):
    def process(self, blob):
        print("New blob:")
        print(blob['CHPrefix'])
        return blob


class Dumper(Module):
    def configure(self):
        self.counter = 0
        self.filenumber = int(time.time())

    def process(self, blob):
        if 'CHData' in blob:
            tag = str(blob['CHPrefix'].tag)
            data = blob['CHData']
            self.dump(data, tag)
        return blob

    def dump(self, data, tag):
        with open('data/{0}_{1}.dat'.format(tag, self.filenumber), 'a+') as f:
            self.counter += 1
            if self.counter % 10000 == 0:
                self.filenumber = int(time.time())
            f.write(data)


pipe = Pipeline()
pipe.attach(
    CHPump,
    host='127.0.0.1',
    port=5553,
    tags='IO_EVT',
    timeout=60 * 60 * 24,
    max_queue=10
)
pipe.attach(CHPrinter)
pipe.attach(Dumper)
pipe.drain()



# Filename: clb.py
"""
Pumps for the CLB data formats.

"""

from io import BytesIO
import struct
from struct import unpack
from collections import namedtuple
import datetime
import pytz

from km3pipe.core import Pump
from km3pipe.sys import ignored
from km3pipe.logger import get_logger

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"

log = get_logger(__name__)    # pylint: disable=C0103

UTC_TZ = pytz.timezone('UTC')


class CLBPump(Pump):
    """A pump for binary CLB files."""
    def configure(self):
        self.filename = self.get('filename')
        self.cache_enabled = self.get('cache_enabled') or False
        self.packet_positions = []
        self.index = 0

        if self.filename:
            self.open_file(self.filename)
            if self.cache_enabled:
                self.determine_packet_positions()

    def determine_packet_positions(self):
        """Record the file pointer position of each frame"""
        print("Analysing file...")
        self.blob_file.seek(0, 0)
        with ignored(struct.error):
            while True:
                pointer_position = self.blob_file.tell()
                length = struct.unpack('<i', self.blob_file.read(4))[0]
                self.packet_positions.append(pointer_position)
                self.blob_file.seek(length, 1)
        self.blob_file.seek(0, 0)
        print("Found {0} CLB UDP packets.".format(len(self.packet_positions)))

    def seek_to_packet(self, index):
        """Move file pointer to the packet with given index."""
        pointer_position = self.packet_positions[index]
        self.blob_file.seek(pointer_position, 0)

    def next_blob(self):
        """Generate next blob in file"""
        try:
            length = struct.unpack('<i', self.blob_file.read(4))[0]
        except struct.error:
            raise StopIteration
        header = CLBHeader(file_obj=self.blob_file)
        blob = {'CLBHeader': header}
        remaining_length = length - header.size
        pmt_data = []
        pmt_raw_data = self.blob_file.read(remaining_length)
        pmt_raw_data_io = BytesIO(pmt_raw_data)
        for _ in range(int(remaining_length / 6)):
            channel_id, time, tot = struct.unpack(
                '>cic', pmt_raw_data_io.read(6)
            )
            pmt_data.append(PMTData(ord(channel_id), time, ord(tot)))
        blob['PMTData'] = pmt_data
        blob['PMTRawData'] = pmt_raw_data
        return blob

    def get_blob(self, index):
        """Return blob at given index."""
        self.seek_to_packet(index)
        return self.next_blob()

    def process(self, blob):
        blob = self.next_blob()
        return blob

    def __iter__(self):
        return self

    def __next__(self):
        return self.next_blob()

    def finish(self):
        """Clean everything up"""
        self.blob_file.close()


class CLBHeader(object):
    """Wrapper for the CLB Common Header binary format.

    Args:
      file_obj (file): The binary file, where the file pointer is at the
        beginning of the header.

    Attributes:
      size (int): The size of the original DAQ byte representation.

    """
    size = 40

    def __init__(self, byte_data=None, file_obj=None):
        self.data_type = None
        self.run = None
        self.udp_sequence = None
        self.timestamp = None
        self.ns_ticks = None
        self.human_readable_timestamp = None
        self.dom_id = None
        self.dom_status = None
        self.time_valid = None
        self.byte_data = byte_data
        if byte_data:
            self._parse_byte_data(byte_data)
        if file_obj:
            self._parse_file(file_obj)

    def __str__(self):
        # pylint: disable=E1124
        description = (
            "CLBHeader\n"
            "    Data type:    {self.data_type}\n"
            "    Run number:   {self.run}\n"
            "    UDP sequence: {self.udp_sequence}\n"
            "    Time stamp:   {self.timestamp}\n"
            "                  {self.human_readable_timestamp}\n"
            "    Ticks [16ns]: {self.ns_ticks}\n"
            "    DOM ID:       {self.dom_id}\n"
            "    DOM status:   {self.dom_status}\n"
            "".format(self=self)
        )
        return description

    def __insp__(self):
        return self.__str__()

    def _parse_byte_data(self, byte_data):
        """Extract the values from byte string."""
        self.data_type = b''.join(unpack('cccc', byte_data[:4])).decode()
        self.run = unpack('>i', byte_data[4:8])[0]
        self.udp_sequence = unpack('>i', byte_data[8:12])[0]
        self.timestamp, self.ns_ticks = unpack('>II', byte_data[12:20])
        self.dom_id = unpack('>i', byte_data[20:24])[0]

        dom_status_bits = unpack('>I', byte_data[24:28])[0]
        self.dom_status = "{0:032b}".format(dom_status_bits)

        self.human_readable_timestamp = datetime.datetime.fromtimestamp(
            int(self.timestamp), UTC_TZ
        ).strftime('%Y-%m-%d %H:%M:%S')

    def _parse_file(self, file_obj):
        """Directly read from file handler.

        Note:
          This will move the file pointer.

        """
        byte_data = file_obj.read(self.size)
        self.byte_data = byte_data
        self._parse_byte_data(byte_data)


PMTData = namedtuple('PMTData', 'channel_id time tot')



#!/usr/bin/env python
# Filename: jpp.py
# pylint: disable=
"""
Pump for the jpp file read through aanet interface.

"""

import numpy as np

from km3pipe.core import Pump, Blob
from km3pipe.dataclasses import Table
from km3pipe.logger import get_logger

log = get_logger(__name__)    # pylint: disable=C0103

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = ["Thomas Heid", "Giuliano Maggi", "Moritz Lotze"]
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"


class EventPump(Pump):
    """A pump for DAQEvents in JPP files.

    Parameters
    ----------
    filename: str
        Name of the file to open.

    """
    def configure(self):

        try:
            import jppy    # noqa
        except ImportError:
            raise ImportError(
                "\nEither Jpp or jppy could not be found."
                "\nMake sure you source the JPP environmanet "
                "and have jppy installed"
            )

        self.event_index = self.get('index') or 0
        self.filename = self.require('filename')

        self.buf_size = 5000
        self._channel_ids = np.zeros(self.buf_size, dtype='i')
        self._dom_ids = np.zeros(self.buf_size, dtype='i')
        self._times = np.zeros(self.buf_size, dtype='i')
        self._tots = np.zeros(self.buf_size, dtype='i')
        self._triggereds = np.zeros(self.buf_size, dtype='i')

        self.event_reader = jppy.PyJDAQEventReader(self.filename.encode())
        self.blobs = self.blob_generator()
        self._current_blob = Blob()

    def _resize_buffers(self, buf_size):
        log.info("Resizing hit buffers to {}.".format(buf_size))
        self.buf_size = buf_size
        self._channel_ids.resize(buf_size, refcheck=False)
        self._dom_ids.resize(buf_size, refcheck=False)
        self._times.resize(buf_size, refcheck=False)
        self._tots.resize(buf_size, refcheck=False)
        self._triggereds.resize(buf_size, refcheck=False)

    def blob_generator(self):
        while self.event_reader.has_next:
            try:
                yield self.extract_event()
            except IndexError:
                pass

        raise StopIteration

    def extract_event(self):
        blob = self._current_blob
        r = self.event_reader
        r.retrieve_next_event()    # do it at the beginning!

        n = r.number_of_snapshot_hits

        if n > self.buf_size:
            self._resize_buffers(int(n * 3 / 2))

        r.get_hits(
            self._channel_ids, self._dom_ids, self._times, self._tots,
            self._triggereds
        )

        hit_series = Table.from_template({
            'channel_id': self._channel_ids[:n],
            'dom_id': self._dom_ids[:n],
            'time': self._times[:n],
            'tot': self._tots[:n],
            'triggered': self._triggereds[:n],
            'group_id': self.event_index,
        }, 'Hits')

        event_info = Table.from_template({
            'det_id': r.det_id,
            'frame_index': r.frame_index,
            'livetime_sec': 0,
            'mc_id': 0,
            'mc_t': 0,
            'n_events_gen': 0,
            'n_files_gen': 0,
            'overlays': r.overlays,
            'trigger_counter': r.trigger_counter,
            'trigger_mask': r.trigger_mask,
            'utc_nanoseconds': r.utc_nanoseconds,
            'utc_seconds': r.utc_seconds,
            'weight_w1': np.nan,
            'weight_w2': np.nan,
            'weight_w3': np.nan,
            'run_id': 0,
            'group_id': self.event_index,
        }, 'EventInfo')

        self.event_index += 1
        blob['EventInfo'] = event_info
        blob['Hits'] = hit_series
        return blob

    def process(self, blob):
        self._current_blob = blob
        return next(self.blobs)

    def __iter__(self):
        return self

    def __next__(self):
        self._current_blob = next(self.blobs)
        return self._current_blob


class TimeslicePump(Pump):
    """A pump to read and extract timeslices. Currently only hits are read.

    Required Parameters
    -------------------
    filename: str
    stream: str ('L0', 'L1', 'L2', 'SN') default: 'L1'

    """
    def configure(self):
        fname = self.require('filename')
        self.stream = self.get('stream', default='L1')
        self.blobs = self.timeslice_generator()
        try:
            import jppy    # noqa
        except ImportError:
            raise ImportError(
                "\nEither Jpp or jppy could not be found."
                "\nMake sure you source the JPP environmanet "
                "and have jppy installed"
            )
        stream = 'JDAQTimeslice' + self.stream
        self.r = jppy.daqtimeslicereader.PyJDAQTimesliceReader(
            fname.encode(), stream.encode()
        )
        self.n_timeslices = self.r.n_timeslices

        self.buf_size = 5000
        self._channel_ids = np.zeros(self.buf_size, dtype='i')
        self._dom_ids = np.zeros(self.buf_size, dtype='i')
        self._times = np.zeros(self.buf_size, dtype='i')
        self._tots = np.zeros(self.buf_size, dtype='i')
        self._triggereds = np.zeros(self.buf_size, dtype=bool)    # dummy

        self._current_blob = Blob()
        self._hits_blob_key = '{}Hits'.format(
            self.stream if self.stream else 'TS'
        )

    def process(self, blob):
        self._current_blob = blob
        return next(self.blobs)

    def timeslice_generator(self):
        """Uses slice ID as iterator"""
        slice_id = 0
        while slice_id < self.n_timeslices:
            blob = self.get_blob(slice_id)
            yield blob
            slice_id += 1

    def get_blob(self, index):
        """Index is slice ID"""
        blob = self._current_blob
        self.r.retrieve_timeslice(index)
        timeslice_info = Table.from_template({
            'frame_index': self.r.frame_index,
            'slice_id': index,
            'timestamp': self.r.utc_seconds,
            'nanoseconds': self.r.utc_nanoseconds,
            'n_frames': self.r.n_frames,
        }, 'TimesliceInfo')
        hits = self._extract_hits()
        hits.group_id = index
        blob['TimesliceInfo'] = timeslice_info
        blob[self._hits_blob_key] = hits
        return blob

    def _extract_hits(self):
        total_hits = self.r.number_of_hits

        if total_hits > self.buf_size:
            buf_size = int(total_hits * 3 / 2)
            self._resize_buffers(buf_size)

        self.r.get_hits(
            self._channel_ids, self._dom_ids, self._times, self._tots
        )

        group_id = 0 if total_hits > 0 else []

        hits = Table.from_template(
            {
                'channel_id': self._channel_ids[:total_hits],
                'dom_id': self._dom_ids[:total_hits],
                'time': self._times[:total_hits].astype('f8'),
                'tot': self._tots[:total_hits],
        # 'triggered': self._triggereds[:total_hits],  # dummy
                'group_id': group_id,    # slice_id will be set afterwards
            },
            'TimesliceHits'
        )
        return hits

    def _resize_buffers(self, buf_size):
        log.info("Resizing hit buffers to {}.".format(buf_size))
        self.buf_size = buf_size
        self._channel_ids.resize(buf_size, refcheck=False)
        self._dom_ids.resize(buf_size, refcheck=False)
        self._times.resize(buf_size, refcheck=False)
        self._tots.resize(buf_size, refcheck=False)
        self._triggereds.resize(buf_size, refcheck=False)    # dummy

    def get_by_frame_index(self, frame_index):
        blob = Blob()
        self.r.retrieve_timeslice_at_frame_index(frame_index)
        hits = self._extract_hits()
        blob[self._hits_blob_key] = hits
        return blob

    def __len__(self):
        return self.n_timeslices

    def __iter__(self):
        return self

    def __next__(self):
        self._current_blob = next(self.blobs)
        return self._current_blob

    def __getitem__(self, index):
        if isinstance(index, int):
            return self.get_blob(index)
        elif isinstance(index, slice):
            return self._slice_generator(index)
        else:
            raise TypeError("index must be int or slice")

    def _slice_generator(self, index):
        """A simple slice generator for iterations"""
        start, stop, step = index.indices(len(self))
        for i in range(start, stop, step):
            yield self.get_blob(i)


class SummaryslicePump(Pump):
    """Preliminary Summaryslice reader"""
    def configure(self):
        filename = self.require('filename')
        self.blobs = self.summaryslice_generator()
        try:
            from jppy.daqsummaryslicereader import PyJDAQSummarysliceReader
        except ImportError:
            raise ImportError(
                "\nEither Jpp or jppy could not be found."
                "\nMake sure you source the JPP environmanet "
                "and have jppy installed"
            )
        self.r = PyJDAQSummarysliceReader(filename.encode())

    def process(self, blob):
        return next(self.blobs)

    def summaryslice_generator(self):
        slice_id = 0
        while self.r.has_next:
            summary_slice = {}
            self.r.retrieve_next_summaryslice()
            blob = Blob()
            summaryslice_info = Table.from_template({
                'frame_index': self.r.frame_index,
                'slice_id': slice_id,
                'timestamp': self.r.utc_seconds,
                'nanoseconds': self.r.utc_nanoseconds,
                'n_frames': self.r.n_frames,
            }, 'SummarysliceInfo')
            blob['SummarysliceInfo'] = summaryslice_info
            while self.r.has_next_frame:
                rates = np.zeros(31, dtype='f8')
                hrvs = np.zeros(31, dtype='i4')
                fifos = np.zeros(31, dtype='i4')
                self.r.get_rates(rates)
                self.r.get_hrvs(hrvs)
                self.r.get_fifos(fifos)
                summary_slice[self.r.dom_id] = {
                    'rates': rates,
                    'hrvs': hrvs.astype(bool),
                    'fifos': fifos.astype(bool),
                    'n_udp_packets': self.r.number_of_received_packets,
                    'max_sequence_number': self.r.max_sequence_number,
                    'has_udp_trailer': self.r.has_udp_trailer,
                    'high_rate_veto': self.r.high_rate_veto,
                    'fifo_status': self.r.fifo_status,
                }
                self.r.retrieve_next_frame()
            blob['Summaryslice'] = summary_slice
            slice_id += 1
            yield blob

    def __iter__(self):
        return self

    def __next__(self):
        return next(self.blobs)


class FitPump(Pump):
    """A pump for JFit objects in JPP files.

    Parameters
    ----------
    filename: str
        Name of the file to open.
    """
    def configure(self):

        try:
            import jppy    # noqa
        except ImportError:
            raise ImportError(
                "\nEither Jpp or jppy could not be found."
                "\nMake sure you source the JPP environmanet "
                "and have jppy installed"
            )

        self.event_index = self.get('index') or 0
        self.filename = self.require('filename')

        self.buf_size = 50000
        self._pos_xs = np.zeros(self.buf_size, dtype='d')
        self._pos_ys = np.zeros(self.buf_size, dtype='d')
        self._pos_zs = np.zeros(self.buf_size, dtype='d')
        self._dir_xs = np.zeros(self.buf_size, dtype='d')
        self._dir_ys = np.zeros(self.buf_size, dtype='d')
        self._dir_zs = np.zeros(self.buf_size, dtype='d')
        self._ndfs = np.zeros(self.buf_size, dtype='i')
        self._times = np.zeros(self.buf_size, dtype='d')
        self._qualities = np.zeros(self.buf_size, dtype='d')
        self._energies = np.zeros(self.buf_size, dtype='d')

        self.event_reader = jppy.PyJFitReader(self.filename.encode())
        self.blobs = self.blob_generator()

    def _resize_buffers(self, buf_size):
        log.info("Resizing hit buffers to {}.".format(buf_size))
        self.buf_size = buf_size
        self._pos_xs.resize(buf_size, refcheck=False)
        self._pos_ys.resize(buf_size, refcheck=False)
        self._pos_zs.resize(buf_size, refcheck=False)
        self._dir_xs.resize(buf_size, refcheck=False)
        self._dir_ys.resize(buf_size, refcheck=False)
        self._dir_zs.resize(buf_size, refcheck=False)
        self._ndfs.resize(buf_size, refcheck=False)
        self._times.resize(buf_size, refcheck=False)
        self._qualities.resize(buf_size, refcheck=False)
        self._energies.resize(buf_size, refcheck=False)

    def blob_generator(self):
        while self.event_reader.has_next:
            try:
                yield self.extract_event()
            except IndexError:
                pass

        raise StopIteration

    def extract_event(self):
        blob = Blob()
        r = self.event_reader
        r.retrieve_next_event()    # do it at the beginning!

        n = r.n_fits

        if n > self.buf_size:
            self._resize_buffers(int(n * 3 / 2))

        r.get_fits(
            self._pos_xs,
            self._pos_ys,
            self._pos_zs,
            self._dir_xs,
            self._dir_ys,
            self._dir_zs,
            self._ndfs,
            self._times,
            self._qualities,
            self._energies,
        )
        fit_collection = Table({
            'pos_x': self._pos_xs[:n],
            'pos_y': self._pos_ys[:n],
            'pos_z': self._pos_zs[:n],
            'dir_x': self._dir_xs[:n],
            'dir_y': self._dir_ys[:n],
            'dir_z': self._dir_zs[:n],
            'ndf': self._ndfs[:n],
            'time': self._times[:n],
            'quality': self._qualities[:n],
            'energy': self._energies[:n],
        },
                               h5loc='/jfit')
        fit_collection = fit_collection.append_columns(['event_id'],
                                                       [self.event_index])

        # TODO make this into a datastructure

        event_info = Table.from_template({
            'det_id': 0,
            'frame_index': 0,
            'livetime_sec': 0,
            'MC ID': 0,
            'MC time': 0,
            'n_events_gen': 0,
            'n_files_gen': 0,
            'overlays': 0,
            'trigger_counter': 0,
            'trigger_mask': 0,
            'utc_nanoseconds': 0,
            'utc_seconds': 0,
            'weight_w1': np.nan,
            'weight_w2': np.nan,
            'weight_w3': np.nan,
            'run_id': 0,
            'group_id': self.event_index,
        }, 'EventInfo')

        self.event_index += 1
        blob['EventInfo'] = event_info
        blob['JFit'] = fit_collection
        return blob

    def process(self, blob):
        nextblob = next(self.blobs)
        blob.update(nextblob)
        return blob

    def __iter__(self):
        return self

    def __next__(self):
        return next(self.blobs)



#!/usr/bin/env python
# Filename: ptconcat.py
"""
Concatenate HDF5 Files.

Usage:
    ptconcat [options] OUTFILE INFILES...
    ptconcat (-h | --help)
    ptconcat --version

Options:
    -h --help               Show this screen.
    --verbose               Print out more progress. [default: False].
    --debug                 Print out A LOT. [default: False].
    --overwrite             Overwrite output file [default: False].
"""

import os.path

import tables as tb

from km3pipe import version

from km3pipe.logger import get_logger
log = get_logger(__name__)

__author__ = "Moritz Lotze"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal and Moritz Lotze"
__email__ = "mlotze@km3net.de"
__status__ = "Development"


def ptconcat(output_file, input_files, overwrite=False):
    """Concatenate HDF5 Files"""
    filt = tb.Filters(
        complevel=5, shuffle=True, fletcher32=True, complib='zlib'
    )
    out_tabs = {}
    dt_file = input_files[0]
    log.info("Reading data struct '%s'..." % dt_file)
    h5struc = tb.open_file(dt_file, 'r')
    log.info("Opening output file '%s'..." % output_file)
    if overwrite:
        outmode = 'w'
    else:
        outmode = 'a'
    h5out = tb.open_file(output_file, outmode)

    for node in h5struc.walk_nodes('/', classname='Table'):
        path = node._v_pathname
        log.debug(path)
        dtype = node.dtype
        p, n = os.path.split(path)
        out_tabs[path] = h5out.create_table(
            p, n, description=dtype, filters=filt, createparents=True
        )
    h5struc.close()
    for fname in input_files:
        log.info('Reading %s...' % fname)
        h5 = tb.open_file(fname)
        for path, out in out_tabs.items():
            tab = h5.get_node(path)
            out.append(tab[:])
        h5.close()
    h5out.close()


def main():
    from docopt import docopt
    args = docopt(__doc__, version=version)

    infiles = args['INFILES']
    outfile = args['OUTFILE']
    overwrite = bool(args['--overwrite']) or False
    verb = bool(args['--verbose'])
    debug = bool(args['--debug'])
    if verb:
        log.setLevel('INFO')
    if debug:
        log.setLevel('DEBUG')

    ptconcat(outfile, infiles, overwrite=overwrite)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# pylint: disable=locally-disabled,C0111,R0904,C0301,C0103,W0212

import sys

for mod in ['aa', 'ROOT']:
    if mod in sys.modules:
        del sys.modules[mod]

import aa    # noqa
from ROOT import Det, EventFile, TH2D

det = Det("KM3NeT_00000007_02122015_zest_DR_PMT.detx")
EventFile.read_timeslices = True
f = EventFile('small.root')
P = 50000    # 50 micro-seconds
h2 = TH2D("h2", "h2", 200, 0, P, 60, 0, 800)
for i, evt in enumerate(f):
    if i > 100:
        break
    if not i % 20:
        print(f.index, f.evt.hits.size())
    if f.index < 1000:
        continue
    det.apply(evt)

    for h in evt.hits:
        h2.Fill(h.t % P, h.pos.z)
    if f.index == 110:
        break
h2.Draw("colz")



# Filename: test_aanet.py
# pylint: disable=locally-disabled,C0111,R0904,C0301,C0103,W0212
from os.path import join, dirname
import time
import numpy as np

import km3pipe as kp
from km3pipe.testing import TestCase, patch, Mock, skipif, skip
from km3pipe.io.aanet import AanetPump, MetaParser
from km3pipe.core import Pipeline, Module, Pump

try:
    import aa
    NO_AANET = False
except ImportError:
    import sys
    sys.modules['ROOT'] = Mock()
    sys.modules['aa'] = Mock()
    NO_AANET = True
else:
    import ROOT

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2018, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"

TEST_DATA_DIR = join(dirname(__file__), "../../kp-data/test_data")


@skipif(NO_AANET, reason="No aanet environment found.")
class TestAanetPump(TestCase):
    def setUp(self):
        self.fname = join(TEST_DATA_DIR, 'sea_data.root')
        self.pump = AanetPump(filename=self.fname)

    def test_reading_hits(self):
        data = {
            'hit_lengths': [48, 71, 22, 44, 35],
            'tot_sums': [1252, 1948, 562, 1089, 915],
            'mean_times': [
                56379691.1875, 65447948.25352113, 51971419.40909091,
                38274188.06818182, 73676555.94285715
            ],
            'channel_id_sums': [556, 958, 372, 593, 433]
        }

        blob_counter = 0
        for idx, blob in enumerate(self.pump):
            blob_counter += 1
            hits = blob['Hits']
            assert len(hits) == data['hit_lengths'][idx]
            self.assertAlmostEqual(data['mean_times'][idx], np.mean(hits.time))
            assert data['tot_sums'][idx] == np.sum(hits.tot)
            assert data['channel_id_sums'][idx] == np.sum(hits.channel_id)
        assert 5 == blob_counter

    def test_event_info(self):
        data = {
            'run_ids': [4143] * 5,
            'event_ids': [0, 1, 2, 3, 4],
            'frame_indices': [60, 41, 80, 61, 110],
            'mc_ids': [59, 40, 79, 60, 109],
            'timestamps': [
                1551549606, 1551549604, 1551549608, 1551549606, 1551549611
            ],
            'nanoseconds': [0, 100000000, 0, 100000000, 0],
            'trigger_counters': [0, 0, 1, 1, 2],
            'trigger_masks': [22, 22, 6, 22, 6],
            'overlays': [6, 14, 2, 6, 3],
        }
        blob_counter = 0
        for idx, blob in enumerate(self.pump):
            blob_counter += 1
            ei = blob['EventInfo'][0]
            assert data['run_ids'][idx] == ei.run_id
            assert data['event_ids'][idx] == ei.event_id
            assert data['frame_indices'][idx] == ei.frame_index
            assert data['timestamps'][idx] == ei.timestamp
            assert data['trigger_counters'][idx] == ei.trigger_counter
            assert data['overlays'][idx] == ei.overlays
            assert data['trigger_masks'][idx] == ei.trigger_mask
            assert data['nanoseconds'][idx] == ei.nanoseconds
        assert 5 == blob_counter

    def test_aanet_and_hdf5_conformity_with_converted_files(self):
        aanet_pump = AanetPump(filename=join(TEST_DATA_DIR, 'mupage.root'))
        hdf5_pump = kp.io.HDF5Pump(
            filename=join(TEST_DATA_DIR, 'mupage.root.h5')
        )

        blob_counter = 0
        for aanet_blob, hdf5_blob in zip(aanet_pump, hdf5_pump):
            blob_counter += 1
            keys = ['Hits', 'McHits', 'McTracks', 'EventInfo']
            for key in keys:
                aanet_data = aanet_blob[key]
                hdf5_data = hdf5_blob[key]
                assert aanet_data.dtype == hdf5_data.dtype

                for attr in aanet_data.dtype.names:
                    # special case for arrays which are all NaNs
                    if np.all(np.isnan(aanet_data[attr])):
                        assert np.all(np.isnan(hdf5_data[attr]))
                        continue
                    # otherwise just compare them
                    assert np.allclose(aanet_data[attr], hdf5_data[attr])

            aanet_header = aanet_blob['Header']
            hdf5_header = hdf5_blob['Header']
            self.assertTupleEqual(aanet_header.seed, hdf5_header.seed)
            self.assertTupleEqual(aanet_header.norma, hdf5_header.norma)
            self.assertTupleEqual(
                aanet_header.coord_origin, hdf5_header.coord_origin
            )
            self.assertTupleEqual(aanet_header.physics, hdf5_header.physics)
            self.assertTupleEqual(aanet_header.propag, hdf5_header.propag)
            self.assertTupleEqual(aanet_header.livetime, hdf5_header.livetime)
            self.assertTupleEqual(aanet_header.cut_in, hdf5_header.cut_in)
            self.assertTupleEqual(aanet_header.can, hdf5_header.can)
            self.assertTupleEqual(
                aanet_header.seabottom, hdf5_header.seabottom
            )
            self.assertTupleEqual(
                aanet_header.start_run, hdf5_header.start_run
            )
            self.assertTupleEqual(aanet_header.simul, hdf5_header.simul)
        assert 3 == blob_counter

    def test_against_single_file_freeze(self):
        start_time = time.time()

        aanet_pump = AanetPump(filename=join(TEST_DATA_DIR, 'mupage.root'))
        assert 10 > (time.time() - start_time)

    def test_reading_hits_from_multiple_files(self):
        # never mix files like this, but it's OK for a test ;)
        aanet_pump = AanetPump(
            filenames=[
                join(TEST_DATA_DIR, 'sea_data.root'),
                join(TEST_DATA_DIR, 'mupage.root'),
                join(TEST_DATA_DIR, 'sea_data.root')
            ]
        )
        blob_counter = 0
        for blob in aanet_pump:
            blob_counter += 1

        assert 13 == blob_counter

    def test_reading_hits_from_multiple_mupage_files(self):
        aanet_pump = AanetPump(
            filenames=[
                join(TEST_DATA_DIR, 'mupage.root'),
                join(TEST_DATA_DIR, 'mupage.root')
            ]
        )
        blob_counter = 0
        for blob in aanet_pump:
            blob_counter += 1

        assert 6 == blob_counter

    def test_reading_hits_from_multiple_corsika_files(self):
        aanet_pump = AanetPump(
            filenames=[
                join(
                    TEST_DATA_DIR,
                    'mcv5.0.DAT004340.propa.sirene.jte.jchain.aanet.4340.root'
                ),
                join(
                    TEST_DATA_DIR,
                    'mcv5.0.DAT004340.propa.sirene.jte.jchain.aanet.4340.root'
                )
            ]
        )
        blob_counter = 0
        for blob in aanet_pump:
            blob_counter += 1

        assert 6 == blob_counter

    def test_reading_multiple_corsika_files_pipeline(self):
        class Tester(kp.Module):
            def configure(self):
                self.counter = 0

            def process(self, blob):
                self.counter = self.counter + 1
                print(self.counter)
                return blob

            def finish(self):
                assert (self.counter == 6)

        pipe = Pipeline(timeit=False)
        pipe.attach(
            AanetPump,
            filenames=[
                join(
                    TEST_DATA_DIR,
                    'mcv5.0.DAT004340.propa.sirene.jte.jchain.aanet.4340.root'
                ),
                join(
                    TEST_DATA_DIR,
                    'mcv5.0.DAT004340.propa.sirene.jte.jchain.aanet.4340.root'
                )
            ]
        )
        pipe.attach(Tester)
        pipe.drain()


class TestMetaParser(TestCase):
    def test_init(self):
        MetaParser()

    def test_parse_string_of_single_entry(self):
        string = b"A 123\nA 1.2.3\nA KM3NET\nA a\nA b\nA Linux"
        mp = MetaParser()
        mp.parse_string(string)

        assert 1 == len(mp.meta)

        assert b'A' == mp.meta[0]['application_name']
        assert b'123' == mp.meta[0]['revision']
        assert b'A a\nA b' == mp.meta[0]['command']

    def test_parse_string_of_multiple_entries(self):
        string = (
            b"A 123\nA 1.2.3\nA KM3NET\nA a\nA b\nA Linux\n"
            b"B 456\nB 4.5.6\nB KM3NET\nB c\nB Linux"
        )
        mp = MetaParser()
        mp.parse_string(string)

        assert 2 == len(mp.meta)

        assert b'A' == mp.meta[0]['application_name']
        assert b'123' == mp.meta[0]['revision']
        assert b'1.2.3' == mp.meta[0]['root_version']
        assert b'A a\nA b' == mp.meta[0]['command']

        assert b'B' == mp.meta[1]['application_name']
        assert b'456' == mp.meta[1]['revision']
        assert b'4.5.6' == mp.meta[1]['root_version']
        assert b'B c' == mp.meta[1]['command']

    def test_parse_testfile(self):
        fname = join(TEST_DATA_DIR, 'jprintmeta.log')
        with open(fname, 'rb') as fobj:
            string = fobj.read()
        mp = MetaParser()
        mp.parse_string(string)
        assert 7 == len(mp.meta)

        assert b'JEvt' == mp.meta[0]['application_name']
        assert b'9912' == mp.meta[0]['revision']
        assert b'5.34/23' == mp.meta[0]['root_version']
        assert mp.meta[0]['command'].startswith(b'JEvt /pbs/throng/km3net')
        assert mp.meta[0]['command'].endswith(b'2 --!')

        assert b'JEnergy' == mp.meta[1]['application_name']
        assert b'9912' == mp.meta[1]['revision']
        assert b'5.34/23' == mp.meta[1]['root_version']
        assert mp.meta[1]['command'].startswith(b'JEnergy /pbs/throng/km3net')
        assert mp.meta[1]['command'].endswith(b'1 --!')

        assert b'JTriggerEfficiency' == mp.meta[-1]['application_name']
        assert b'8519' == mp.meta[-1]['revision']
        assert b'5.34/23' == mp.meta[-1]['root_version']
        assert mp.meta[-1]['command'].startswith(
            b'JTriggerEfficiency /pbs/throng/km3net'
        )
        assert mp.meta[-1]['command'].endswith(b'326 --!')

    def test_get_table(self):
        string = (
            b"A 123\nA 1.2.3\nA KM3NET\nA a\nA b\nA Linux\n"
            b"B 456\nB 4.5.6\nB KM3NET\nB c\nB Linux"
        )
        mp = MetaParser()
        mp.parse_string(string)
        tab = mp.get_table()
        assert 2 == len(tab)
        assert '/meta' == tab.h5loc
        assert tab.h5singleton
        assert b'B c' == tab['command'][1]
        assert np.dtype('S7') == tab['command'].dtype

    def test_empty_string(self):
        mp = MetaParser()
        mp.parse_string('')
        assert mp.get_table() is None



#!usr/bin/env python
# -*- coding: utf-8 -*-
# Filename: stats.py
# pylint: disable=C0103
"""
Statistics.
"""
import numpy as np
from scipy.stats import rv_continuous

from .math import log_b
from .logger import get_logger
from .tools import zero_pad

__author__ = "Moritz Lotze"
__copyright__ = "Copyright 2017, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Moritz Lotze"
__email__ = "mlotze@km3net.de"
__status__ = "Development"

log = get_logger(__name__)    # pylint: disable=C0103


class loguniform(rv_continuous):
    """Loguniform Distributon"""
    def __init__(self, low=0.1, high=1, base=10, *args, **kwargs):
        super(loguniform, self).__init__(*args, **kwargs)
        self._log_low = log_b(low, base=base)
        self._log_high = log_b(high, base=base)
        self._base_of_log = base

    def _rvs(self, *args, **kwargs):
        # `rvs(size=foo, *args)` does argcheck etc, and sets `self._size`
        return np.power(
            self._base_of_log,
            np.random.uniform(self._log_low, self._log_high, self._size)
        )


class rv_kde(rv_continuous):
    """Create a `scipy.stats.rv_continuous` instance from a (gaussian) KDE.

    Uses the KDE implementation from sklearn.

    Automatic bandwidth,  either from the statsmodels or scipy implementation.
    """
    def __init__(
        self, data, bw=None, bw_method=None, bw_statsmodels=False, **kde_args
    ):
        from sklearn.neighbors import KernelDensity
        from sklearn.utils import check_array
        data = check_array(data, order='C')
        if bw is None:
            if bw_statsmodels:
                bw = self._bandwidth_statsmodels(data)
            else:
                bw = self._bandwidth_scipy(data, bw_method)
        self._kde = KernelDensity(bandwidth=bw, **kde_args).fit(data)
        self.bandwidth = bw
        self.kernel = self._kde.get_params()['kernel']
        super(rv_kde, self).__init__(name='KDE')

    @staticmethod
    def _bandwidth_statsmodels(sample):
        from statsmodels.nonparametric.kernel_density import KDEMultivariate
        # all continuous
        vt = sample.shape[1] * 'c'
        skde = KDEMultivariate(sample, var_type=vt)
        bw = skde.bw
        return bw

    @staticmethod
    def _bandwidth_scipy(sample, bw_method=None):
        from scipy.stats import gaussian_kde
        # sklearn expects switched shape versus scipy
        sample = sample.T
        gkde = gaussian_kde(sample, bw_method=bw_method)
        f = gkde.covariance_factor()
        bw = f * sample.std()
        return bw

    def pdf(self, x):
        # we implement `pdf` instead of `_pdf`, since
        # otherwise scipy performs reshaping of `x` which messes
        # things up for sklearn -- we wanna reshape ourselves!
        from sklearn.utils import check_array
        x = check_array(x, order='C')
        log_pdf = self._kde.score_samples(x)
        pdf = np.exp(log_pdf)
        return pdf

    def rvs(self, *args, **kwargs):
        """Draw Random Variates.

        Parameters
        ----------
        size: int, optional (default=1)
        random_state_: optional (default=None)
        """
        # TODO REVERSE THIS FUCK PYTHON2
        size = kwargs.pop('size', 1)
        random_state = kwargs.pop('size', None)
        # don't ask me why it uses `self._size`
        return self._kde.sample(n_samples=size, random_state=random_state)


def mad(v):
    """MAD -- Median absolute deviation. More robust than standard deviation.
    """
    return np.median(np.abs(v - np.median(v)))


def mad_std(v):
    """Robust estimate of standard deviation using the MAD.

    Lifted from astropy.stats."""
    MAD = mad(v)
    return MAD * 1.482602218505602


def drop_zero_variance(df):
    """Remove columns from dataframe with zero variance."""
    return df.copy().loc[:, df.var() != 0].copy()


def param_names(scipy_dist):
    """Get names of fit parameters from a ``scipy.rv_*`` distribution."""
    if not isinstance(scipy_dist, rv_continuous):
        raise TypeError
    names = ['loc', 'scale']
    if scipy_dist.shapes is not None:
        names += scipy_dist.shapes.split()
    return names


def perc(arr, p=95, **kwargs):
    """Create symmetric percentiles, with ``p`` coverage."""
    offset = (100 - p) / 2
    return np.percentile(arr, (offset, 100 - offset), **kwargs)


def resample_1d(arr, n_out=None, random_state=None):
    """Resample an array, with replacement.

    Parameters
    ==========
    arr: np.ndarray
        The array is resampled along the first axis.
    n_out: int, optional
        Number of samples to return. If not specified,
        return ``len(arr)`` samples.
    """
    if random_state is None:
        random_state = np.random.RandomState()
    arr = np.atleast_1d(arr)
    n = len(arr)
    if n_out is None:
        n_out = n
    idx = random_state.randint(0, n, size=n)
    return arr[idx]


def bootstrap_params(rv_cont, data, n_iter=5, **kwargs):
    """Bootstrap the fit params of a distribution.

    Parameters
    ==========
    rv_cont: scipy.stats.rv_continuous instance
        The distribution which to fit.
    data: array-like, 1d
        The data on which to fit.
    n_iter: int [default=10]
        Number of bootstrap iterations.
    """
    fit_res = []
    for _ in range(n_iter):
        params = rv_cont.fit(resample_1d(data, **kwargs))
        fit_res.append(params)
    fit_res = np.array(fit_res)
    return fit_res


def param_describe(params, quant=95, axis=0):
    """Get mean + quantile range from bootstrapped params."""
    par = np.mean(params, axis=axis)
    lo, up = perc(quant)
    p_up = np.percentile(params, up, axis=axis)
    p_lo = np.percentile(params, lo, axis=axis)
    return par, p_lo, p_up


def bootstrap_fit(
    rv_cont, data, n_iter=10, quant=95, print_params=True, **kwargs
):
    """Bootstrap a distribution fit + get confidence intervals for the params.

    Parameters
    ==========
    rv_cont: scipy.stats.rv_continuous instance
        The distribution which to fit.
    data: array-like, 1d
        The data on which to fit.
    n_iter: int [default=10]
        Number of bootstrap iterations.
    quant: int [default=95]
        percentile of the confidence limits (default is 95, i.e. 2.5%-97.5%)
    print_params: bool [default=True]
        Print a fit summary.
    """
    fit_params = bootstrap_params(rv_cont, data, n_iter)
    par, lo, up = param_describe(fit_params, quant=quant)
    names = param_names(rv_cont)
    maxlen = max([len(s) for s in names])
    print("--------------")
    print(rv_cont.name)
    print("--------------")
    for i, name in enumerate(names):
        print(
            "{nam:>{fill}}: {mean:+.3f} ∈ "
            "[{lo:+.3f}, {up:+.3f}] ({q}%)".format(
                nam=name,
                fill=maxlen,
                mean=par[i],
                lo=lo[i],
                up=up[i],
                q=quant
            )
        )
    out = {
        'mean': par,
        'lower limit': lo,
        'upper limit': up,
    }
    return out


class hist2d(rv_continuous):
    """Simple implementation of a 2d histogram."""
    def __init__(self, H2D, *args, **kwargs):
        H, xlims, ylims = H2D
        self.H = H
        self.xlims = xlims
        self.ylims = ylims
        self.H /= self.integral
        self.xcenters = bincenters(xlims)
        self.ycenters = bincenters(ylims)
        super(hist2d, self).__init__(*args, **kwargs)

    @property
    def H_pad(self):
        return zero_pad(self.H)

    def _pdf(self, X):
        X = np.atleast_2d(X)
        x = X[:, 0]
        y = X[:, 1]
        return self.H_pad[np.searchsorted(self.xlims, x, side='right'),
                          np.searchsorted(self.ylims, y, side='right'),
                          ]

    @property
    def integral(self):
        return (self.H * self.area).sum()

    @property
    def area(self):
        xwidths = np.diff(self.xlims)
        ywidths = np.diff(self.ylims)
        area = np.outer(xwidths, ywidths)
        return area


def bincenters(bins):
    """Bincenters, assuming they are all equally spaced."""
    bins = np.atleast_1d(bins)
    return 0.5 * (bins[1:] + bins[:-1])



# Filename: test_jpp.py
# pylint: disable=locally-disabled,C0111,R0904,C0301,C0103,W0212
from km3pipe.testing import TestCase, surrogate, patch

from km3pipe.io.jpp import EventPump, TimeslicePump, SummaryslicePump, FitPump

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2018, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"


class TestEventPump(TestCase):
    @surrogate('jppy.PyJDAQEventReader')
    @patch('jppy.PyJDAQEventReader')
    def test_init(self, reader_mock):
        filename = 'a.root'
        EventPump(filename=filename)
        reader_mock.assert_called_with(filename.encode())

    @surrogate('jppy.PyJDAQEventReader')
    @patch('jppy.PyJDAQEventReader')
    def test_resize_buffers(self, reader_mock):
        filename = 'a.root'
        pump = EventPump(filename=filename)
        assert len(pump._channel_ids) == pump.buf_size
        new_buf_size = 8000
        pump._resize_buffers(new_buf_size)
        assert pump.buf_size == new_buf_size
        assert len(pump._channel_ids) == new_buf_size


class TestTimeslicePump(TestCase):
    @surrogate('jppy.daqtimeslicereader.PyJDAQTimesliceReader')
    @patch('jppy.daqtimeslicereader.PyJDAQTimesliceReader')
    def test_init(self, reader_mock):
        filename = 'a.root'
        TimeslicePump(filename=filename)
        reader_mock.assert_called_with(filename.encode(), b'JDAQTimesliceL1')

    @surrogate('jppy.daqtimeslicereader.PyJDAQTimesliceReader')
    @patch('jppy.daqtimeslicereader.PyJDAQTimesliceReader')
    def test_init_with_specific_stream(self, reader_mock):
        filename = 'a.root'
        TimeslicePump(filename=filename, stream='L0')
        reader_mock.assert_called_with(filename.encode(), b'JDAQTimesliceL0')
        TimeslicePump(filename=filename, stream='SN')
        reader_mock.assert_called_with(filename.encode(), b'JDAQTimesliceSN')

    @surrogate('jppy.daqtimeslicereader.PyJDAQTimesliceReader')
    @patch('jppy.daqtimeslicereader.PyJDAQTimesliceReader')
    def test_resize_buffers(self, reader_mock):
        filename = 'a.root'
        pump = TimeslicePump(filename=filename)
        assert len(pump._channel_ids) == pump.buf_size
        new_buf_size = 8000
        pump._resize_buffers(new_buf_size)
        assert pump.buf_size == new_buf_size
        assert len(pump._channel_ids) == new_buf_size


class TestSummaryslicePump(TestCase):
    @surrogate('jppy.daqsummaryslicereader.PyJDAQSummarysliceReader')
    @patch('jppy.daqsummaryslicereader.PyJDAQSummarysliceReader')
    def test_init(self, reader_mock):
        filename = 'a.root'
        SummaryslicePump(filename=filename)
        reader_mock.assert_called_with(filename.encode())


class TestFitPump(TestCase):
    @surrogate('jppy.PyJFitReader')
    @patch('jppy.PyJFitReader')
    def test_init(self, reader_mock):
        filename = 'a.root'
        FitPump(filename=filename)
        reader_mock.assert_called_with(filename.encode())

    @surrogate('jppy.PyJFitReader')
    @patch('jppy.PyJFitReader')
    def test_resize_buffers(self, reader_mock):
        filename = 'a.root'
        pump = FitPump(filename=filename)
        assert len(pump._pos_xs) == pump.buf_size
        new_buf_size = 8000
        pump._resize_buffers(new_buf_size)
        assert pump.buf_size == new_buf_size
        assert len(pump._pos_xs) == new_buf_size
